#include <GL/gl.h>
#include <GL/glx.h>
#include <forms.h>
#include <stdlib.h>
#include <stdio.h>
#include <malloc.h>
#include <string.h>
#include "xfpp3d.h"

FL_IO_CALLBACK handle_stdin(int fd,FD_form *ui);
void           badstream_exit(char *s);
void           getinput(void);
double         clamped(double val,double min,double max);
void           tumble(void);

/* History: gl.c demo from xforms distribution and pp3d.c */

int       Trace     = 0;
char     *InputFile = (char *) NULL;
int       EOF_stdin = 0;
int       Detached_stdin = 0;
int       N_idle_cb = 0;     
int       Freq_idle_cb = 100;     

FD_form   *ui;

/*  For status info ... */
char          Buffer[1024];

/*  Defines frustum for projection ... */

double        ProjectionBbox[] = { -1.0, 1.0, -1.0, 1.0, 5.0, 15.0 };

/*  For tumbling rotations ... */

int           Tumble = 1;

double        Xrot  = 0.0;
double        Xstep = 0.1;
double        Yrot  = 0.0;
double        Ystep = 0.2;
double        Zrot  = 0.0;
double        Zstep = 0.3;

double        TumbleFactor = 1.1;

/* For Zoom-in/Zoom-out ... */

double        ZoomFactor = 1.1;
double        Zoom = 1.0;

int           DrawOrigin = 1;

int           Ltrace = 0;

int           Win_w,       Win_h;

double        Win_Xsize = 500;
double        Win_Ysize = 500;

int           Init = 0, 
              Done = 0;

double        Xmin = -2.0, 
              Ymin = -2.0, 
              Xmax =  2.0,
              Ymax =  2.0;

double        BaseSize =  5.0;
double        MinSize  =  3.0;
double        MaxSize  = 20.0;

int           Np = 0;
double        TimeLoc;
double       *Size   = (double *) NULL;
double       *Coords = (double *) NULL;

int           Nt = 0;

int           Pause = 0;
double        DSize = 1.2;
double        Trans = 0.05;
char          Win_Title[1024];

int           Dbuff = 1;
int           AntiA = 1;

double        R[] = { 1.0, 1.0, 0.0, 0.0, 1.0, 0.0, 1.0 },
              G[] = { 1.0, 0.0, 1.0, 0.0, 1.0, 1.0, 0.0 },
              B[] = { 1.0, 0.0, 0.0, 1.0, 0.0, 1.0, 1.0 };
int           Ncol = sizeof(R) / sizeof(double);

double        Dummy;

/*--------------------------------------------------------------------*/
/* Dumps error message to stderr and exits                            */
/*--------------------------------------------------------------------*/
void badstream_exit(char *s) {
   fprintf(stderr,"Error reading %s\n",s);
   exit(0);
}

/*--------------------------------------------------------------------*/
/* Processes input                                                    */
/*--------------------------------------------------------------------*/
void getinput(void) {
   int    i;

   if( Ltrace ) {
      fprintf(stderr,"getinput: In routine: Done <%d> Pause <%d> Init <%d>\n",
              Done,Pause,Init);
   }
   if( Done || Pause ) {
      idle_cb(0, 0);
      return;
   }

   if( Init == 0 ) {
      Init = 1;
      idle_cb(0, 0);
      return;
   }
   if( Init == 1 ) {
/*-------------------------------------------------------------------
      Initialization: Read number of particles and particle sizes,
      allocate storage for particle coordinates and sizes.
  -------------------------------------------------------------------*/
      if( scanf("%d",&Np) == 1 )  {
         if( Ltrace ) {
            fprintf(stderr,"getinput: %d particles\n",Np);
         }
         Size = (double *) malloc(Np * sizeof(double));
         Coords = (double *) malloc(3 * Np * sizeof(double));
         for( i = 0; i < Np; i++ ) {
            if( scanf("%lf",Size + i) == 1 ) {
               if( Ltrace ) {
                  fprintf(stderr,"getinput: Particle %d  Size %g\n",
                          i+1,Size[i]);
               }
            } else {
               badstream_exit("particle sizes");
            }
         }
      } else {
         badstream_exit("number of particles");
      }
      Init = 2;
   } 
/*-------------------------------------------------------------------
   Read new time, particle positions and post redisplay request.
  -------------------------------------------------------------------*/
   if( scanf("%lf",&TimeLoc) == 1 ) {
      if( Ltrace ) fprintf(stderr,"getinput: Time %g\n",TimeLoc);
      for( i = 0; i < Np; i++ ) {
         if( scanf("%lf %lf %lf",Coords + 3*i, Coords + 3*i + 1,Coords + 3*i + 2) ==
             3 ) {
            if( Ltrace ) {
               fprintf(stderr,"getinput: Particle %d  Pos (%g,%g,%g)\n",
                       i,Coords[3*i],Coords[3*i+1],Coords[3*i+2]);
            }
         }
      }
      Nt++;
      if( Nt == 1 ) {
         Pause = 1;  
         fl_set_button(ui->pause_button,Pause);
/*       Need to trigger expose event ... */
      }
   } else {
      Done = 1;
      if(  Tumble ) {
/*       glutSetWindowTitle("<ESC> exits, t stops tumble, h displays help");  */
      }  else {
/*       glutSetWindowTitle("<ESC> exits, t starts tumble, h displays help");  */
      }
   }
   idle_cb(0,0);
}

/*--------------------------------------------------------------------*/
/* Displays particles ... most of the OpenGL graphics calls are       */
/* in this routine.                                                   */
/*--------------------------------------------------------------------*/
int N_display = 0;
void display(void) {
   int    i;
   int    ltrace;

   if( Ltrace ) {
      fprintf(stderr,"display: In routine: Np <%d>\n",Np);
   }
   N_display++;
   ltrace = Ltrace ? !(N_display % 100) : 0;
   if( ltrace ) fprintf(stderr,"display: In routine\n");

   fl_activate_glcanvas(ui->canvas);

   glMatrixMode(GL_MODELVIEW);

   glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

   glPushMatrix();
   glTranslatef( 0.0, 0.0, -0.5 * (ProjectionBbox[4] + ProjectionBbox[5]));
   if( ltrace ) {
      fprintf(stderr,"display: Rotation angles: %g %g %g\n",Xrot,Yrot,Zrot);
   }
   glRotated( Xrot, 1.0, 0.0, 0.0 );
   glRotated( Yrot, 0.0, 1.0, 0.0 );
   glRotated( Zrot, 0.0, 0.0, 1.0 );

   glScaled( Zoom, Zoom, Zoom );

   if( Np ) {

      for( i = 0; i < Np; i++ ) {
         int ic = i % Ncol;
         if( ltrace ) {
            fprintf(stderr,"display: Particle %d  Size %g  Pos (%g,%g,%g)\n", 
                    i+1,BaseSize*Size[i],Coords[3*i],Coords[3*i+1],Coords[3*i+2]);
         }
         glColor3d(R[1],G[1],B[1]);
         glPointSize((GLfloat) clamped(BaseSize * Size[i],MinSize,MaxSize));
         glBegin(GL_POINTS);
            glVertex3d(Coords[3*i],Coords[3*i+1],Coords[3*i+2]);
         glEnd();
      } 
   }
   if( DrawOrigin ) {
      glColor3d(R[0],G[0],B[0]);
      glPointSize((GLfloat) (BaseSize * 1));
      glBegin(GL_POINTS);
         glVertex3d(0.0,0.0,0.0);
      glEnd();
   }

   glXSwapBuffers(fl_display, fl_get_canvas_id(ui->canvas));  

   glFlush();
   glPopMatrix();
}

void tumble(void) {

   if( !Tumble ) return;

   if( Ltrace ) {
      fprintf(stderr,"tumble:  Rotation angles: %g %g %g\n",Xrot,Yrot,Zrot);
   }
   Xrot += Xstep;
   Yrot += Ystep;
   Zrot += Zstep;

   if (Xrot>=360.0) {
      Xrot = Xrot - 360;
   }
   else if (Yrot>=360.0) {
      Yrot = Yrot - 360;
   }
   else if (Zrot>=360.0) {
      Zrot = Zrot - 360;
   }
}


int buttonpress_cb(FL_OBJECT *ob, Window win, int w, int h, 
                   XEvent *xev, void *ud)  {
   return 0;
}

void zoom_in_button_cb(FL_OBJECT *ob, long val) {
   Zoom *= ZoomFactor;
   idle_cb();
}

void zoom_out_button_cb(FL_OBJECT *ob, long val) {
   Zoom /= ZoomFactor;
   idle_cb();
}

void exit_button_cb(FL_OBJECT *ob, long val) {
   if( fl_show_question("Really Exit?",1) ) exit(1);
}

void tumble_button_cb(FL_OBJECT *ob, long val) {
   Tumble = !Tumble;
   idle_cb();
}

void tumble_slower_button_cb(FL_OBJECT *ob, long val) {
   Xstep /= TumbleFactor;
   Ystep /= TumbleFactor;
   Zstep /= TumbleFactor;
   idle_cb();
}

void tumble_faster_button_cb(FL_OBJECT *ob, long val) {
   Xstep *= TumbleFactor;
   Ystep *= TumbleFactor;
   Zstep *= TumbleFactor;
   idle_cb();
}

void pause_button_cb(FL_OBJECT *ob, long val) {
   Pause = !Pause;
   idle_cb();
}

void origin_button_cb(FL_OBJECT *ob, long val) {
   DrawOrigin = !DrawOrigin;
   idle_cb();
}

void bbox_button_cb(FL_OBJECT *on, long val) {
   fprintf(stderr,"bbox_button_cb: Not implemented yet\n");
   idle_cb();
}

int gl_init(void) {
    if( AntiA ) glEnable(GL_POINT_SMOOTH); 

    glClearColor(0.0,0.0,0.0,0.0);

    glMatrixMode( GL_PROJECTION );
    glLoadIdentity();
    glFrustum(ProjectionBbox[0],ProjectionBbox[1],
              ProjectionBbox[2],ProjectionBbox[3],
              ProjectionBbox[4],ProjectionBbox[5]);

    glMatrixMode( GL_MODELVIEW ); 
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    glXSwapBuffers(fl_display, fl_get_canvas_id(ui->canvas));  

    return 0;
}

int canvas_expose(FL_OBJECT *ob, Window win, int w, int h, XEvent *xev, void *ud)
{
    glViewport(0,0, (GLint)w, (GLint)h);
    gl_init();

    /* refresh */
    idle_cb(0, 0);
    return 0;
}


int idle_cb(XEvent *ev, void *data) {
   if(!ui->form->visible  || ! ui->canvas->visible) {
      return 0;
   }
   N_idle_cb++;
   if( Trace && !(N_idle_cb % Freq_idle_cb) ) {
       fprintf(stderr,"idle_cb: In routine %d\n",N_idle_cb);
   }
   if( EOF_stdin && !Detached_stdin ) {
      fl_remove_io_callback(0,FL_READ,(FL_IO_CALLBACK) handle_stdin);
      Detached_stdin = 1;
      if( Trace ) {
         fprintf(stderr,"idle_cb:  Removed handle_stdin callback\n");
          sleep(2);
      }
   }

/* Set status line ... */
   if( Done ) {
      sprintf(Buffer,"Done Reading (%d steps)",Nt);
   } else {
      if( Pause ) {
         sprintf(Buffer,"Reading/Paused (Step %d)",Nt);
      } else {
         sprintf(Buffer,"Reading (Step %d)",Nt);
      }
   }
   fl_set_object_label(ui->status_text,Buffer);

/* Set time ... */
   sprintf(Buffer,"T = %g",TimeLoc);
   fl_set_object_label(ui->time_text,Buffer);

/* Set number of particles ... */
   sprintf(Buffer,"%d particles",Np);
   fl_set_object_label(ui->np_text,Buffer);

   display();
   tumble();
  
   return 0;
}

void input_file_cb(FL_OBJECT *the, long arg) {
   if( InputFile ) {
      free(InputFile);
   }
   InputFile = strdup(fl_get_input(the));
   if( Trace ) {
      fprintf(stderr,"input_file_cb: Input File: <%s>\n",InputFile);
   }
}

FL_IO_CALLBACK handle_stdin(int fd,FD_form *ui) {
   getinput();

   return 0;
}


FL_IO_CALLBACK handle_stdin_old(int fd,FD_form *ui) {
   int    val;

   if( EOF_stdin) return 0;

   if( Trace ) {
      fprintf(stderr,"handle_stdin:  In routine\n");
   }
   switch( fscanf(stdin,"%d",&val) ) {
   case EOF:
      EOF_stdin = 1;
      break;
   case 0:
      fprintf(stderr,"handle_stdin: Error reading stdin\n");
      break;
   case 1:
      fprintf(stderr,"handle_stdin: Read <%d>\n",val);
      break;
   default:
      fprintf(stderr,"handle_stdin Unexpected case\n");
      break;
   }

   return 0;
}

/*--------------------------------------------------------------------*/
/* Returns val clamped to min ... max                                 */
/*--------------------------------------------------------------------*/
double clamped(double val,double min,double max) {
   if(        val < min ) {
      return min;
   } else if( val > max ) {
      return max;
   } else {
      return val;
   }
} 


int main(int argc, char *argv[]) {

   fl_initialize(&argc, argv, "xfpp3d", 0, 0);
   ui = create_form_form();
   fl_show_form(ui->form,
                FL_PLACE_CENTER | FL_FREE_SIZE, 
                FL_FULLBORDER,
                "xfpp3d");

   fl_add_canvas_handler(ui->canvas, Expose, canvas_expose, 0);
   fl_add_canvas_handler(ui->canvas, ButtonPress, buttonpress_cb, 0);
   fl_set_idle_delta(1);
   fl_set_idle_callback(idle_cb, 0);

   fl_add_io_callback(0,FL_READ,(FL_IO_CALLBACK) handle_stdin,ui); 
             
   /* geometry stuff */
   fl_set_form_minsize(ui->form, 340, 280);

   gl_init();
   fl_set_button(ui->tumble_button,Tumble);
   fl_set_button(ui->pause_button,Pause);
   fl_set_button(ui->origin_button,DrawOrigin);
   fl_do_forms();
   fl_finish();

   return 0;
}
